import mujoco as mj
from mujoco.glfw import glfw
import numpy as np
import os
import utility as ram
from RNEA import RNEA #Recursive Newton Euler Algorithm
import time

xml_path = 'planar_3R.xml' #xml file (assumes this is in the same folder as this file)
simend = 0.5 #simulation time
print_camera_config = 0 #set to 1 to print camera config
                        #this is useful for initializing view of the model)

# For callback functions
button_left = False
button_middle = False
button_right = False
lastx = 0
lasty = 0

def init_controller(model,data):
    #initialize the controller here. This function is called once, in the beginning
    pass

def controller(model, data):
    #put the controller here. This function is called inside the simulation.
    pass

def keyboard(window, key, scancode, act, mods):
    if act == glfw.PRESS and key == glfw.KEY_BACKSPACE:
        mj.mj_resetData(model, data)
        mj.mj_forward(model, data)

def mouse_button(window, button, act, mods):
    # update button state
    global button_left
    global button_middle
    global button_right

    button_left = (glfw.get_mouse_button(
        window, glfw.MOUSE_BUTTON_LEFT) == glfw.PRESS)
    button_middle = (glfw.get_mouse_button(
        window, glfw.MOUSE_BUTTON_MIDDLE) == glfw.PRESS)
    button_right = (glfw.get_mouse_button(
        window, glfw.MOUSE_BUTTON_RIGHT) == glfw.PRESS)

    # update mouse position
    glfw.get_cursor_pos(window)

def mouse_move(window, xpos, ypos):
    # compute mouse displacement, save
    global lastx
    global lasty
    global button_left
    global button_middle
    global button_right

    dx = xpos - lastx
    dy = ypos - lasty
    lastx = xpos
    lasty = ypos

    # no buttons down: nothing to do
    if (not button_left) and (not button_middle) and (not button_right):
        return

    # get current window size
    width, height = glfw.get_window_size(window)

    # get shift key state
    PRESS_LEFT_SHIFT = glfw.get_key(
        window, glfw.KEY_LEFT_SHIFT) == glfw.PRESS
    PRESS_RIGHT_SHIFT = glfw.get_key(
        window, glfw.KEY_RIGHT_SHIFT) == glfw.PRESS
    mod_shift = (PRESS_LEFT_SHIFT or PRESS_RIGHT_SHIFT)

    # determine action based on mouse button
    if button_right:
        if mod_shift:
            action = mj.mjtMouse.mjMOUSE_MOVE_H
        else:
            action = mj.mjtMouse.mjMOUSE_MOVE_V
    elif button_left:
        if mod_shift:
            action = mj.mjtMouse.mjMOUSE_ROTATE_H
        else:
            action = mj.mjtMouse.mjMOUSE_ROTATE_V
    else:
        action = mj.mjtMouse.mjMOUSE_ZOOM

    mj.mjv_moveCamera(model, action, dx/height,
                      dy/height, scene, cam)

def scroll(window, xoffset, yoffset):
    action = mj.mjtMouse.mjMOUSE_ZOOM
    mj.mjv_moveCamera(model, action, 0.0, -0.05 *
                      yoffset, scene, cam)

#get the full path
dirname = os.path.dirname(__file__)
abspath = os.path.join(dirname + "/" + xml_path)
xml_path = abspath

# MuJoCo data structures
model = mj.MjModel.from_xml_path(xml_path)  # MuJoCo model
data = mj.MjData(model)                # MuJoCo data
cam = mj.MjvCamera()                        # Abstract camera
opt = mj.MjvOption()                        # visualization options

# Init GLFW, create window, make OpenGL context current, request v-sync
glfw.init()
window = glfw.create_window(1200, 900, "Demo", None, None)
glfw.make_context_current(window)
glfw.swap_interval(1)

# initialize visualization data structures
mj.mjv_defaultCamera(cam)
mj.mjv_defaultOption(opt)
scene = mj.MjvScene(model, maxgeom=10000)
context = mj.MjrContext(model, mj.mjtFontScale.mjFONTSCALE_150.value)

# install GLFW mouse and keyboard callbacks
glfw.set_key_callback(window, keyboard)
glfw.set_cursor_pos_callback(window, mouse_move)
glfw.set_mouse_button_callback(window, mouse_button)
glfw.set_scroll_callback(window, scroll)

# Example on how to set camera configuration
# cam.azimuth = 90
# cam.elevation = -45
# cam.distance = 2
# cam.lookat = np.array([0.0, 0.0, 0])
cam.azimuth = 89.8231032550546 ; cam.elevation = -88.84857804659835 ; cam.distance =  6.025299868401033
cam.lookat =np.array([ 0.0 , 0.0 , 0.0 ])

#get the key
key_id = model.key("home").id
key_qpos = model.key_qpos[key_id]  #Access qpos
key_ctrl = model.key_ctrl[key_id] #Access ctrl

#see https://mujoco.readthedocs.io/en/latest/python.html
#https://mujoco.readthedocs.io/en/latest/APIreference/APItypes.html#model
#method 1
key_id = model.key("home").id
key_qpos = model.key_qpos[key_id]

#method2
# key_qpos = model.key("home").qpos

model.opt.gravity = np.array([0, -9.81, 0])

q = np.array([0.5, -1.2, 1.57])
u = np.array([1.23, 6.54, 1.89])

#initialize the controller
init_controller(model,data)

#set the controller
mj.set_mjcb_control(controller)

while not glfw.window_should_close(window):
    time_prev = data.time

    if (data.time<0.01):
        #q = key_qpos.copy()
        data.qpos = q.copy()
        data.qvel = u.copy()

    while (data.time - time_prev < 1.0/60.0):

        #kinematics
        data.time += 0.2 #model.opt.timestep
        data.qpos = q.copy()
        mj.mj_forward(model,data)

        #tic = time.time()  # Start timer

        C_py,G_py,CG_py,M_py = RNEA(q,u)

        #toc = time.time()  # End timer
        #print(f"Elapsed time: {toc - tic:.4f} seconds")



        #M*qddot + qfrc_bias = tau
        #M*qddot + C + G = tau
        #so C+G = qfrc_bias
        print("Gravity and Coriolis")
        qfrc_bias = np.zeros((model.nv,1))
        for i in range(0,model.nv):
            qfrc_bias[i]=data.qfrc_bias[i]
        ram.print_matrix(qfrc_bias,4,"CG_mj")
        ram.print_matrix(CG_py,4,"CG_py")
        ram.print_matrix(CG_py-qfrc_bias,6,"CG_py-CG_mj ")


        print("Mass matrix")
        M_mj = np.zeros((model.nv,model.nv ))
        mj.mj_fullM(model, M_mj, data.qM)
        ram.print_matrix(M_mj,6,"M_mj")
        ram.print_matrix(M_py,6,"M_py")

        ram.print_matrix(M_py-M_mj,6,"M_py-M_mj")



    if (data.time>=simend):
        break;

    # get framebuffer viewport
    viewport_width, viewport_height = glfw.get_framebuffer_size(
        window)
    viewport = mj.MjrRect(0, 0, viewport_width, viewport_height)

    #print camera configuration (help to initialize the view)
    if (print_camera_config==1):
        print('cam.azimuth =',cam.azimuth,';','cam.elevation =',cam.elevation,';','cam.distance = ',cam.distance)
        print('cam.lookat =np.array([',cam.lookat[0],',',cam.lookat[1],',',cam.lookat[2],'])')

    # Update scene and render
    mj.mjv_updateScene(model, data, opt, None, cam,
                       mj.mjtCatBit.mjCAT_ALL.value, scene)
    mj.mjr_render(viewport, scene, context)

    # swap OpenGL buffers (blocking call due to v-sync)
    glfw.swap_buffers(window)

    # process pending GUI events, call GLFW callbacks
    glfw.poll_events()

glfw.terminate()
